
/****** addivec.c *******************************************************/
/*									*/
/*   Included in the module is a global combine program based on the	*/
/*   algorithm from "On the Efficiency of Global Combine Algorithm for	*/
/*   2-D Meshes with Wormhole Routing" by M. Barnell et al.		*/
/*									*/
/*   The algorithm implemented in this program is Recursive Halving/	*/
/*   Bidirectional Exchange Hybrid. It only applys for the cases that	*/
/*   both number of nodes and size of arrays are power of 2.		*/
/*									*/
/***************************************************** Shin 4/4/93 ******/

#include <stdio.h>
#include "parti.h"
#include "bisec.h"

#define	 DefaultSize	512	/* threshold size for recursive halving */

/*
 * Global sum of integer arrays
 */
void
addivec (level, buf, ndata, temp)
int   level;			/* level of the cube */
int   *buf;			/* buffer of input array */
int   ndata;			/* size of input array */
int   *temp;			/* temporary array */

{
   int   start;                 /* start position of buffer */
   int   curndata;               /* current size of array */
   int   msgndata;               /* current size of array */
   int   halflevel;             /* level of recursive halving */
   int   channel;               /* communicate channel */
   int   mask;
   int   partner;		/* partner processor id */
   int   *ndatalist;		/* size of array for each level */
   int   i, j, k, n;

   /* initialize variables */
   curndata = ndata;
   halflevel = 0;
   start = 0;			/* start position of buffer */

   /* allocalte memory for list of ndata */
   if (ndata > DefaultSize) {
     ndatalist = (int *) malloc ((level+2)*sizeof(int));
     ndatalist[0] = ndata;
   }

   /* perform global combine opertion */
   for (i = 0; i < level; i++) {
      /* find the partner */
      mask = 1 << i;
      channel = PARTI_mynode & mask;		/* inconnection channel */
      partner = PARTI_mynode ^ mask;		/* partner id */

      if (curndata <= DefaultSize) {
         /* when the size of array is less than DefaultSize
            the algorithm switched to Bidirectional Exchange */
	 /* send array to partner */
	 if (channel) {
	    PARTI_csend (EXCHODD+i, &buf[start], curndata*sizeof(int), partner, PARTI_pid);
	    PARTI_crecv (EXCHEVEN+i, &temp[start], curndata*sizeof(int), partner, PARTI_pid);
	 }
	 else {
	    PARTI_csend (EXCHEVEN+i, &buf[start], curndata*sizeof(int), partner, PARTI_pid);
	    PARTI_crecv (EXCHODD+i, &temp[start], curndata*sizeof(int), partner, PARTI_pid);
         }

	 /* combine the buffer and tempoary */
	 for (j = 0, n = start; j < curndata; j++, n++)
	    buf[n] += temp[n];
	
	 /* continue the next loop */
	 continue;
      }

      /* when the size of array is large than DefaultSize
	 then perform the Recurive Halving algorithm */
      msgndata = curndata;
      if (channel) {
         /* calculate the size of sending and receiving messages */
         msgndata >>= 1;		/* half of the array */
         curndata = (curndata & 1)? msgndata+1: msgndata;

	 /* send the lower half and receive upper half of array */
	 PARTI_csend (REVODD+i, &buf[start], msgndata*sizeof(int), partner, PARTI_pid);
	 PARTI_crecv (REVEVEN+i, &temp[start+msgndata], curndata*sizeof(int), partner, PARTI_pid);

	 /* combine the lower half of array */
	 for (j = 0, n = start+msgndata; j < curndata; j++, n++)
	    buf[n] += temp[n];
	
	 /* update the start position */
	 start += msgndata;
      }
      else {
	 /* calculate the size of sending and receiving messages */
	 msgndata >>= 1;
         if (curndata & 1) msgndata++;
	 curndata >>= 1;

	 /* send the upper half and receive lower half of array */
	 PARTI_csend (REVEVEN+i, &buf[start+curndata], msgndata*sizeof(int), partner, PARTI_pid);
	 PARTI_crecv (REVODD+i, &temp[start], curndata*sizeof(int), partner, PARTI_pid);

	 /* combine the lower half of array */
	 for (j = 0, n = start; j < curndata; j++, n++)
	    buf[n] += temp[n];
      }
      halflevel++;		/* increase redistributing level */
      ndatalist[halflevel] = curndata;
   }

   /* redistribute the global sums of subarrays
      when Recursive Halving is applied */
   for (i = halflevel; i > 0; i--) {
      mask = 1 << (i-1);		/* mask for communication */
      channel = PARTI_mynode & mask;
      partner = PARTI_mynode ^ mask;

      /* send the result to partner */
      if (channel) {
         /* calculate the new sizes of sending and receiving messages */
         curndata = msgndata = ndatalist[i];
         if (ndatalist[i-1] & 1)
           msgndata--;

	 /* update the starting position of array */
	 start -= msgndata;

	 /* send the upper half and receive the lower half of array */
	 PARTI_csend (DISTODD+i, &buf[start+msgndata], curndata*sizeof(int), partner, PARTI_pid);
	 PARTI_crecv (DISTEVEN+i, &buf[start], msgndata*sizeof(int), partner, PARTI_pid);
      }
      else {
         /* calculate the new sizes of sending and receiving messages */
         curndata = msgndata = ndatalist[i];
         if (ndatalist[i-1] & 1)
           msgndata++;

	 /* send the lower half and receive upper half of array */
	 PARTI_csend (DISTEVEN+i, &buf[start], curndata*sizeof(int), partner, PARTI_pid);
	 PARTI_crecv (DISTODD+i, &buf[start+curndata], msgndata*sizeof(int), partner, PARTI_pid);
      }
   }

   /* free allocated memory */
   if (ndata > DefaultSize)
     free (ndatalist);
}

/*
 *  wrapper function for Fortran
 */
void
addivec_ (level, buf, ndata, temp)
int   *level;			/* level of the cube */
int   *buf;			/* buffer of input array */
int   *ndata;			/* size of input array */
int   *temp;			/* temporary array */

{
   addivec (*level, buf, *ndata, temp);
}

/*
 * Global sum of double arrays
 */
void
adddvec (level, buf, ndata, temp)
int    level;			/* level of the cube */
double *buf;			/* buffer of input array */
int    ndata;			/* size of input array */
double *temp;			/* temporary array */

{
   int   start;                 /* start position of buffer */
   int   curndata;               /* current size of array */
   int   msgndata;               /* current size of array */
   int   halflevel;             /* level of recursive halving */
   int   channel;               /* communicate channel */
   int   mask;
   int   partner;		/* partner processor id */
   int   *ndatalist;		/* size of array for each level */
   int   i, j, k, n;

   /* initialize variables */
   curndata = ndata;
   halflevel = 0;
   start = 0;			/* start position of buffer */

   /* allocalte memory for list of ndata */
   if (ndata > DefaultSize) {
     ndatalist = (int *) malloc ((level+2)*sizeof(int));
     ndatalist[0] = ndata;
   }

   /* perform global combine opertion */
   for (i = 0; i < level; i++) {
      /* find the partner */
      mask = 1 << i;
      channel = PARTI_mynode & mask;		/* inconnection channel */
      partner = PARTI_mynode ^ mask;		/* partner id */

      if (curndata <= DefaultSize) {
         /* when the size of array is less than DefaultSize
            the algorithm switched to Bidirectional Exchange */
	 /* send array to partner */
	 if (channel) {
	    PARTI_csend (EXCHODD+i, &buf[start], curndata*sizeof(double), partner, PARTI_pid);
	    PARTI_crecv (EXCHEVEN+i, &temp[start], curndata*sizeof(double), partner, PARTI_pid);
	 }
	 else {
	    PARTI_csend (EXCHEVEN+i, &buf[start], curndata*sizeof(double), partner, PARTI_pid);
	    PARTI_crecv (EXCHODD+i, &temp[start], curndata*sizeof(double), partner, PARTI_pid);
         }

	 /* combine the buffer and tempoary */
	 for (j = 0, n = start; j < curndata; j++, n++)
	    buf[n] += temp[n];
	
	 /* continue the next loop */
	 continue;
      }

      /* when the size of array is large than DefaultSize
	 then perform the Recurive Halving algorithm */
      msgndata = curndata;
      if (channel) {
         /* calculate the size of sending and receiving messages */
         msgndata >>= 1;                /* half of the array */
         curndata = (curndata & 1)? msgndata+1: msgndata;

	 /* send the lower half and receive upper half of array */
	 PARTI_csend (REVODD+i, &buf[start], msgndata*sizeof(double), partner, PARTI_pid);
	 PARTI_crecv (REVEVEN+i, &temp[start+msgndata], curndata*sizeof(double), partner, PARTI_pid);

	 /* combine the lower half of array */
	 for (j = 0, n = start+msgndata; j < curndata; j++, n++)
	    buf[n] += temp[n];
	
	 /* update the start position */
	 start += msgndata;
      }
      else {
         /* calculate the size of sending and receiving messages */
	 msgndata >>= 1;
         if (curndata & 1) msgndata++;
	 curndata >>= 1;

	 /* send the upper half and receive lower half of array */
	 PARTI_csend (REVEVEN+i, &buf[start+curndata], msgndata*sizeof(double), partner, PARTI_pid);
	 PARTI_crecv (REVODD+i, &temp[start], curndata*sizeof(double), partner, PARTI_pid);

	 /* combine the lower half of array */
	 for (j = 0, n = start; j < curndata; j++, n++)
	    buf[n] += temp[n];
      }
      halflevel++;              /* increase redistributing level */
      ndatalist[halflevel] = curndata;
   }

   /* redistribute the global sums of subarrays
      when Recursive Halving is applied */
   for (i = halflevel; i > 0; i--) {
      mask = 1 << (i-1);			/* mask for communication */
      channel = PARTI_mynode & mask;
      partner = PARTI_mynode ^ mask;

      /* send the result to partner */
      if (channel) {
         /* calculate the new sizes of sending and receiving messages */
         curndata = msgndata = ndatalist[i];
         if (ndatalist[i-1] & 1)
           msgndata--;

	 /* update the starting position of array */
	 start -= msgndata;

	 /* send the upper half and receive the lower half of array */
	 PARTI_csend (DISTODD+i, &buf[start+msgndata], curndata*sizeof(double), partner, PARTI_pid);
	 PARTI_crecv (DISTEVEN+i, &buf[start], msgndata*sizeof(double), partner, PARTI_pid);
      }
      else {
         /* calculate the new sizes of sending and receiving messages */
         curndata = msgndata = ndatalist[i];
         if (ndatalist[i-1] & 1)
           msgndata++;

	 /* send the lower half and receive upper half of array */
	 PARTI_csend (DISTEVEN+i, &buf[start], curndata*sizeof(double), partner, PARTI_pid);
	 PARTI_crecv (DISTODD+i, &buf[start+curndata], msgndata*sizeof(double), partner, PARTI_pid);
      }
   }

   /* free allocated memory */
   if (ndata > DefaultSize)
     free (ndatalist);
}

adddvec_ (level, buf, ndata, temp)
int    *level;			/* level of the cube */
double *buf;			/* buffer of input array */
int    *ndata;			/* size of input array */
double *temp;			/* temporary array */

{
   adddvec (*level, buf, *ndata, temp);
}

/*
 * global sum of Val arrays
 */
void
addsvec (level, buf, ndata, temp)
int   level;			/* level of the cube */
Val   *buf;			/* buffer of input array */
int   ndata;			/* size of input array */
Val   *temp;			/* temporary array */

{
   int   start;                 /* start position of buffer */
   int   curndata;               /* current size of array */
   int   msgndata;               /* current size of array */
   int   halflevel;             /* level of recursive halving */
   int   channel;               /* communicate channel */
   int   mask;
   int   partner;		/* partner processor id */
   int   *ndatalist;            /* size of array for each level */
   int   i, j, k, n;

   /* initialize variables */
   curndata = ndata;
   halflevel = 0;
   start = 0;			/* start position of buffer */

   /* allocalte memory for list of ndata */
   if (ndata > DefaultSize) {
     ndatalist = (int *) malloc ((level+2)*sizeof(int));
     ndatalist[0] = ndata;
   }

   /* perform global combine opertion */
   for (i = 0; i < level; i++) {
      /* find the partner */
      mask = 1 << i;
      channel = PARTI_mynode & mask;		/* inconnection channel */
      partner = PARTI_mynode ^ mask;		/* partner id */

      if (curndata <= DefaultSize) {
         /* when the size of array is less than DefaultSize
            the algorithm switched to Bidirectional Exchange */
	 /* send array to partner */
	 if (channel) {
	    PARTI_csend (EXCHODD+i, &buf[start], curndata*sizeof(Val), partner, PARTI_pid);
	    PARTI_crecv (EXCHEVEN+i, &temp[start], curndata*sizeof(Val), partner, PARTI_pid);
	 }
	 else {
	    PARTI_csend (EXCHEVEN+i, &buf[start], curndata*sizeof(Val), partner, PARTI_pid);
	    PARTI_crecv (EXCHODD+i, &temp[start], curndata*sizeof(Val), partner, PARTI_pid);
         }

	 /* combine the buffer and tempoary */
	 for (j = 0, n = start; j < curndata; j++, n++) {
	    buf[n].val += temp[n].val;
	    buf[n].index += temp[n].index;
	 }
	
	 /* continue the next loop */
	 continue;
      }

      /* when the size of array is large than DefaultSize
	 then perform the Recurive Halving algorithm */
      msgndata = curndata;
      if (channel) {
         /* calculate the size of sending and receiving messages */
         msgndata >>= 1;                /* half of the array */
         curndata = (curndata & 1)? msgndata+1: msgndata;

	 /* send the lower half and receive upper half of array */
	 PARTI_csend (REVODD+i, &buf[start], msgndata*sizeof(Val), partner, PARTI_pid);
	 PARTI_crecv (REVEVEN+i, &temp[start+msgndata], curndata*sizeof(Val), partner, PARTI_pid);

	 /* combine the lower half of array */
	 for (j = 0, n = start+msgndata; j < curndata; j++, n++) {
	    buf[n].val += temp[n].val;
	    buf[n].index += temp[n].index;
	 }
	
	 /* update the start position */
	 start += msgndata;
      }
      else {
         /* calculate the size of sending and receiving messages */
	 msgndata >>= 1;
         if (curndata & 1) msgndata++;
	 curndata >>= 1;

	 /* send the upper half and receive lower half of array */
	 PARTI_csend (REVEVEN+i, &buf[start+curndata], msgndata*sizeof(Val), partner, PARTI_pid);
	 PARTI_crecv (REVODD+i, &temp[start], curndata*sizeof(Val), partner, PARTI_pid);

	 /* combine the lower half of array */
	 for (j = 0, n = start; j < curndata; j++, n++) {
	    buf[n].val += temp[n].val;
	    buf[n].index += temp[n].index;
	 }
      }
      halflevel++;              /* increase redistributing level */
      ndatalist[halflevel] = curndata;
   }

   /* redistribute the global sums of subarrays
      when Recursive Halving is applied */
   for (i = halflevel; i > 0; i--) {
      mask = 1 << (i-1);			/* mask for communication */
      channel = PARTI_mynode & mask;
      partner = PARTI_mynode ^ mask;

      /* send the result to partner */
      if (channel) {
         /* calculate the new sizes of sending and receiving messages */
         curndata = msgndata = ndatalist[i];
         if (ndatalist[i-1] & 1)
           msgndata--;

	 /* update the starting position of array */
	 start -= msgndata;

	 /* send the upper half and receive the lower half of array */
	 PARTI_csend (DISTODD+i, &buf[start+msgndata], curndata*sizeof(Val), partner, PARTI_pid);
	 PARTI_crecv (DISTEVEN+i, &buf[start], msgndata*sizeof(Val), partner, PARTI_pid);
      }
      else {
         /* calculate the new sizes of sending and receiving messages */
         curndata = msgndata = ndatalist[i];
         if (ndatalist[i-1] & 1)
           msgndata++;

	 /* send the lower half and receive upper half of array */
	 PARTI_csend (DISTEVEN+i, &buf[start], curndata*sizeof(Val), partner, PARTI_pid);
	 PARTI_crecv (DISTODD+i, &buf[start+curndata], msgndata*sizeof(Val), partner, PARTI_pid);
      }
   }

   /* free allocated memory */
   if (ndata > DefaultSize)
     free (ndatalist);
}

/*
 * Global sum of double arrays
 * -- Number of float point operations is returned
 */
int
adddvec_flop (level, buf, ndata, temp)
int    level;			/* level of the cube */
double *buf;			/* buffer of input array */
int    ndata;			/* size of input array */
double *temp;			/* temporary array */

{
   int   start;                 /* start position of buffer */
   int   curndata;               /* current size of array */
   int   msgndata;               /* current size of array */
   int   halflevel;             /* level of recursive halving */
   int   channel;               /* communicate channel */
   int   mask;
   int   partner;		/* partner processor id */
   int   *ndatalist;		/* size of array for each level */
   int   i, j, k, n;
   int   noflop = 0;		/* number of float point operations */

   /* initialize variables */
   curndata = ndata;
   halflevel = 0;
   start = 0;			/* start position of buffer */

   /* allocalte memory for list of ndata */
   if (ndata > DefaultSize) {
     ndatalist = (int *) malloc ((level+2)*sizeof(int));
     ndatalist[0] = ndata;
   }

   /* perform global combine opertion */
   for (i = 0; i < level; i++) {
      /* find the partner */
      mask = 1 << i;
      channel = PARTI_mynode & mask;		/* inconnection channel */
      partner = PARTI_mynode ^ mask;		/* partner id */

      if (curndata <= DefaultSize) {
         /* when the size of array is less than DefaultSize
            the algorithm switched to Bidirectional Exchange */
	 /* send array to partner */
	 if (channel) {
	    PARTI_csend (EXCHODD+i, &buf[start], curndata*sizeof(double), partner, PARTI_pid);
	    PARTI_crecv (EXCHEVEN+i, &temp[start], curndata*sizeof(double), partner, PARTI_pid);
	 }
	 else {
	    PARTI_csend (EXCHEVEN+i, &buf[start], curndata*sizeof(double), partner, PARTI_pid);
	    PARTI_crecv (EXCHODD+i, &temp[start], curndata*sizeof(double), partner, PARTI_pid);
         }

	 /* combine the buffer and tempoary */
	 for (j = 0, n = start; j < curndata; j++, n++)
	    buf[n] += temp[n];
	
         /* count the number of float point operations */
         noflop += curndata;

	 /* continue the next loop */
	 continue;
      }

      /* when the size of array is large than DefaultSize
	 then perform the Recurive Halving algorithm */
      msgndata = curndata;
      if (channel) {
         /* calculate the size of sending and receiving messages */
         msgndata >>= 1;                /* half of the array */
         curndata = (curndata & 1)? msgndata+1: msgndata;

	 /* send the lower half and receive upper half of array */
	 PARTI_csend (REVODD+i, &buf[start], msgndata*sizeof(double), partner, PARTI_pid);
	 PARTI_crecv (REVEVEN+i, &temp[start+msgndata], curndata*sizeof(double), partner, PARTI_pid);

	 /* combine the lower half of array */
	 for (j = 0, n = start+msgndata; j < curndata; j++, n++)
	    buf[n] += temp[n];
	
         /* count the number of float point operations */
         noflop += curndata;

	 /* update the start position */
	 start += msgndata;
      }
      else {
         /* calculate the size of sending and receiving messages */
	 msgndata >>= 1;
         if (curndata & 1) msgndata++;
	 curndata >>= 1;

	 /* send the upper half and receive lower half of array */
	 PARTI_csend (REVEVEN+i, &buf[start+curndata], msgndata*sizeof(double), partner, PARTI_pid);
	 PARTI_crecv (REVODD+i, &temp[start], curndata*sizeof(double), partner, PARTI_pid);

	 /* combine the lower half of array */
	 for (j = 0, n = start; j < curndata; j++, n++)
	    buf[n] += temp[n];

         /* count the number of float point operations */
         noflop += curndata;
      }
      halflevel++;              /* increase redistributing level */
      ndatalist[halflevel] = curndata;
   }

   /* redistribute the global sums of subarrays
      when Recursive Halving is applied */
   for (i = halflevel; i > 0; i--) {
      mask = 1 << (i-1);			/* mask for communication */
      channel = PARTI_mynode & mask;
      partner = PARTI_mynode ^ mask;

      /* send the result to partner */
      if (channel) {
         /* calculate the new sizes of sending and receiving messages */
         curndata = msgndata = ndatalist[i];
         if (ndatalist[i-1] & 1)
           msgndata--;

	 /* update the starting position of array */
	 start -= msgndata;

	 /* send the upper half and receive the lower half of array */
	 PARTI_csend (DISTODD+i, &buf[start+msgndata], curndata*sizeof(double), partner, PARTI_pid);
	 PARTI_crecv (DISTEVEN+i, &buf[start], msgndata*sizeof(double), partner, PARTI_pid);
      }
      else {
         /* calculate the new sizes of sending and receiving messages */
         curndata = msgndata = ndatalist[i];
         if (ndatalist[i-1] & 1)
           msgndata++;

	 /* send the lower half and receive upper half of array */
	 PARTI_csend (DISTEVEN+i, &buf[start], curndata*sizeof(double), partner, PARTI_pid);
	 PARTI_crecv (DISTODD+i, &buf[start+curndata], msgndata*sizeof(double), partner, PARTI_pid);
      }
   }

   /* free allocated memory */
   if (ndata > DefaultSize)
     free (ndatalist);

   /* return the number of float point operations */
   return (noflop);
}

int
adddvec_flop_ (level, buf, ndata, temp)
int    *level;			/* level of the cube */
double *buf;			/* buffer of input array */
int    *ndata;			/* size of input array */
double *temp;			/* temporary array */

{
   return (adddvec_flop (*level, buf, *ndata, temp));
}
